/* defines.h	- Definitions of things needed in all C files
 *
 * $Revision: 2.41 $
 * $Date: 1999/01/05 23:46:04 $
 */

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <errno.h>

/* Defines for ifdef'd code -- define them in the Makefile */

/* #define DEBUG			adds in debugging code */
/* #define STREAM_BUFFERING		converts reads/writes to stdio calls */
/* #define ZERO_MEMORY			zeros all of process memory before
					it starts to run */
/* #define NATIVES			allows native binaries and PDP-11
					binaries in the filespace */
					

/* Optimisation defines */
#ifndef INLINE
# define INLINE
#endif

/* Special defines to enable/disable certain
 * functionality. These are added as required
 * to port to new platforms. Please send in new
 * defines, please!
 */

#if __FreeBSD__ < 3
#define NO_GETPGID
#endif


/* Type definitions for PDP data types. You may have to
 * modify the #if conditions to suit your own system.
 * If you do make changes to the #if's, then I would
 * be very happy to include them.
 *
 * Warning: I will probably switch from u_int_*  to
 * u_int_* soon. Let me know if this will cause any
 * problems.	Warren wkt@cs.adfa.edu.au
 */

#if __FreeBSD__ < 2
typedef char int8_t;
typedef short int16_t;
typedef long int32_t;
#endif
#if __FreeBSD__ < 2
typedef unsigned char u_int8_t;
typedef unsigned short u_int16_t;
typedef unsigned long u_int32_t;
#endif

/* I should probably put 6 assert lines here to double-check
 * that the 6 types above are the correct sizes in bits
 */

/* Instruction Table for Fast Decode. */

typedef void (*_itab) ();


/* PDP processor defines. */

#define SP	6		/* stack pointer */
#define PC	7		/* program counter */
#define CC_NBIT 010		/* Positions of the flags in the */
#define CC_ZBIT 04		/* processor status word */
#define CC_VBIT 02
#define CC_CBIT 01

#define PDP_MEM_SIZE	65536
#define MAX_ARGS	200


/* Global variables. */

extern u_int16_t regs[8];	/* general registers */
extern u_int16_t ir;		/* current instruction register */
extern int CC_N;		/* The processor status word is represented */
extern int CC_Z;		/* by these four values. On some */
extern int CC_V;		/* architectures, you may get a performance */
extern int CC_C;

extern u_int16_t ea_addr;	/* stored address for dest modifying insts */
extern _itab itab[];
extern u_int8_t *ispace, *dspace;
extern u_int16_t dwrite_base;	/* Lowest addr where dspace writes can occur */
extern _itab itab[1024];	/* Main list of instruction functions */

extern u_int16_t dstword;	/* These globals are used in the effective */
extern u_int16_t srcword;	/* address calculations, mainly to save */
extern u_int16_t tmpword;	/* parameter passing overheads in */
extern u_int8_t	dstbyte;	/* function calls */
extern u_int8_t	srcbyte;
extern u_int8_t	tmpbyte;

extern char realfilename[500];	/* These 2 are used on filename translation */
extern char *rfn;		/* to the APOUT_ROOT */
extern int sig_arrived;		/* Indicates if a signal has arrived */
extern int Argc, Envc;		/* Arguments passed to new process */
extern char *Argv[MAX_ARGS], *Envp[MAX_ARGS];

				/* 2.11BSD overlay stuff */
extern u_int32_t ov_changes;	/* Number of overlay changes */

#ifdef DEBUG
				/* Debugging flags */
extern inst_debug,		/* Print a line before each instruction */
    trap_debug,			/* Print details of each trap */
    jsr_debug,			/* Print out each jsr */
    fp_debug;			/* Print out each floating-point instruction */
extern FILE *dbg_file;		/* Debugging output file */
extern char *progname;		/* The program's name - used in debugging */
#endif


/* Inline defines. */

#define CLR_CC_V()	CC_V=0
#define CLR_CC_C()	CC_C=0
#define CLR_CC_Z()	CC_Z=0
#define CLR_CC_N()	CC_N=0
#define CLR_CC_ALL()	CC_V=CC_C=CC_Z=CC_N=0

#define SET_CC_V()	CC_V=1
#define SET_CC_C()	CC_C=1
#define SET_CC_Z()	CC_Z=1
#define SET_CC_N()	CC_N=1

#define SRC_MODE	(( ir & 07000 ) >> 9 )
#define SRC_REG		(( ir & 0700 ) >> 6 )
#define DST_MODE	(( ir & 070 ) >> 3 )
#define DST_REG		( ir & 07 )

#define LSBIT	1		/* least significant bit */

#define MPI	0077777		/* most positive integer */
#define MNI	0100000		/* most negative integer */
#define NEG_1	0177777		/* negative one */
#define SIGN	0100000		/* sign bit */
#define CARRY	0200000		/* set if carry out */

#define MPI_B	0177		/* most positive integer (byte) */
#define MNI_B	0200		/* most negative integer (byte) */
#define NEG_1_B 0377		/* negative one (byte) */
#define SIGN_B	0200		/* sign bit (byte) */
#define CARRY_B 0400		/* set if carry out (byte) */

#define LOW16( data )	(( data ) & 0177777 )	/* mask the lower 16 bits */
#define LOW8( data )	(( data ) & 0377 )	/* mask the lower 8 bits */

#define CHG_CC_N( d )	if ((d) & SIGN ) \
					SET_CC_N(); \
				else \
					CLR_CC_N()

#define CHGB_CC_N( d )	if ((d) & SIGN_B ) \
				SET_CC_N(); \
			else \
				CLR_CC_N()

#define CHG_CC_Z( d )	if ( d ) \
					CLR_CC_Z(); \
				else \
					SET_CC_Z()

#define CHGB_CC_Z( d )	if ( LOW8( d )) \
				CLR_CC_Z(); \
			else \
				SET_CC_Z()

#define CHG_CC_C( d )	if ((d) & CARRY ) \
					SET_CC_C(); \
				else \
					CLR_CC_C()

#define CHG_CC_IC( d )	if ((d) & CARRY ) \
					CLR_CC_C(); \
				else \
					SET_CC_C()

#define CHGB_CC_IC( d ) if ((d) & CARRY_B ) \
				CLR_CC_C(); \
			else \
				SET_CC_C()

#define CHG_CC_V( d1, d2, d3 )	\
				if ((( d1 & SIGN ) == ( d2 & SIGN )) \
				&& (( d1 & SIGN ) != ( d3 & SIGN ))) \
					SET_CC_V(); \
				else \
					CLR_CC_V()

#define CHG_CC_VC( d1, d2, d3 ) \
				if ((( d1 & SIGN ) != ( d2 & SIGN )) \
				&& (( d2 & SIGN ) == ( d3 & SIGN ))) \
					SET_CC_V(); \
				else \
					CLR_CC_V()

#define CHG_CC_VS( d1, d2, d3 ) \
				if ((( d1 & SIGN ) != ( d2 & SIGN )) \
				&& (( d1 & SIGN ) == ( d3 & SIGN ))) \
					SET_CC_V(); \
				else \
					CLR_CC_V()

#define CHGB_CC_V( d1, d2, d3 ) \
				if ((( d1 & SIGN_B ) == ( d2 & SIGN_B )) \
				&& (( d1 & SIGN_B ) != ( d3 & SIGN_B ))) \
					SET_CC_V(); \
				else \
					CLR_CC_V()

#define CHGB_CC_VC(d1,d2,d3)	\
				if ((( d1 & SIGN_B ) != ( d2 & SIGN_B )) \
				&& (( d2 & SIGN_B ) == ( d3 & SIGN_B ))) \
					SET_CC_V(); \
				else \
					CLR_CC_V()

#define CHG_CC_V_XOR_C_N()	\
				if ((( CC_C ) && \
				   ( CC_N )) \
				|| ((!( CC_C )) && \
				   ( ! ( CC_N )))) \
					CLR_CC_V(); \
				else \
					SET_CC_V()


/* Macros to read and write loctions in
 * main memory.
 */

extern u_int16_t *adptr;

/* lli_word() - Load a word from the given ispace logical address. */
#define lli_word(addr, word) \
	{ adptr= (u_int16_t *)&(ispace[addr]); word= *adptr; }

/* ll_word() - Load a word from the given logical address. */
#define ll_word(addr, word) \
	{ adptr= (u_int16_t *)&(dspace[addr]); word= *adptr; }

/* sl_word() - Store a word at the given logical address. */
#define sl_word(addr, word) \
	{ if (addr < dwrite_base) seg_fault(); \
	  adptr= (u_int16_t *)&(dspace[addr]); *adptr= word; }

/* lli_byte() - Load a byte from the given logical ispace address. */
#define lli_byte(addr, byte) \
    byte = ispace[addr];

/* ll_byte() - Load a byte from the given logical address. */
#define ll_byte(addr, byte) \
    byte = dspace[addr];

/* sl_byte() - Store a byte at the given logical address. */
#define sl_byte(addr, byte) \
	{ if (addr < dwrite_base) seg_fault(); \
	  dspace[addr]= byte; }

#define copylong(to,from) \
	buf = (char *) &(to); buf2 = (char *) &(from); \
	buf[0]=buf2[2]; buf[1]=buf2[3]; buf[2]=buf2[0]; buf[3]=buf2[1]



/* Function prototypes */

#ifdef __STDC__
#define P(s) s
#else
#define P(s) ()
#endif

/* aout.c */
int load_a_out P((const char *file, int v6exec));
#ifdef EMU211
void do_bsd_overlay P((void));
#endif

/* branch.c */
void br P((void));
void blos P((void));
void bge P((void));
void blt P((void));
void ble P((void));
void bgt P((void));
void jmp P((void));
void jsr P((void));
void rts P((void));
void bne P((void));
void beq P((void));
void bpl P((void));
void bmi P((void));
void bhi P((void));
void bvc P((void));
void bvs P((void));
void bcc P((void));
void bcs P((void));
void scc P((void));
void ccc P((void));
void sob P((void));
void mfps P((void));
void mtps P((void));
void mfpi P((void));
void mtpi P((void));
void ash P((void));
void mul P((void));
void divide P((void));
void ashc P((void));
void xor P((void));

/* double.c */
void mov P((void));
void cmp P((void));
void add P((void));
void sub P((void));
void bit P((void));
void bic P((void));
void bis P((void));
void movb P((void));
void cmpb P((void));
void bitb P((void));
void bicb P((void));
void bisb P((void));

/* cpu.c */
void run P((void));
void sim_init P((void));
void bus_error P((void));
void seg_fault P((void));
void waiti P((void));
void halt P((void));
void iot P((void));
void emt P((void));
void bpt P((void));
void fis P((void));
void illegal P((void));
void not_impl P((void));
void mark P((void));
void mfpd P((void));
void mtpd P((void));
void trap P((void));
void bad_FP_reg P((void));

/* ea.c */
void load_ea P((void));
void pop P((void));
void push P((void));
void loadb_dst P((void));
void loadb_src P((void));
void storeb_dst P((void));
void storeb_dst_2 P((void));
void loadp_dst P((void));
void storep_dst P((void));
void load_src P((void));
void store_dst P((void));
void load_dst P((void));
void store_dst_2 P((void));

/* fp.c */
void fpset P((void));
void ldf P((void));
void stf P((void));
void clrf P((void));
void addf P((void));
void subf P((void));
void negf P((void));
void absf P((void));
void mulf P((void));
void moddf P((void));
void divf P((void));
void cmpf P((void));
void tstf P((void));
void ldfps P((void));
void stfps P((void));
void lcdif P((void));
void stcfi P((void));
void stexp P((void));
void stcdf P((void));
void ldcdf P((void));
void stst P((void));
void ldexpp P((void));

/* itab.c */
void dositab0 P((void));
void dositab1 P((void));

/* main.c */
int main P((int argc, char **argv));
char * xlate_filename P((char *name));

/* single.c */
void adc P((void));
void asl P((void));
void asr P((void));
void clr P((void));
void com P((void));
void dec P((void));
void inc P((void));
void neg P((void));
void rol P((void));
void ror P((void));
void sbc P((void));
void swabi P((void));
void sxt P((void));
void tst P((void));
void tstb P((void));
void aslb P((void));
void asrb P((void));
void clrb P((void));
void comb P((void));
void decb P((void));
void incb P((void));
void negb P((void));
void rolb P((void));
void rorb P((void));
void adcb P((void));
void sbcb P((void));

/* v7trap.c */
void v7trap P((void));

/* bsdtrap.c */
#ifdef EMU211
void bsdtrap P((void));
#endif

#undef P
